"""babyspider.crawler - Functions to crawl a web domain and download resources"""

import logging
import importlib
import math
import time
import warnings
from collections.abc import Callable
from datetime import datetime
from pathlib import Path
from urllib.error import URLError
from urllib.parse import urljoin, urlparse, urlunparse
from urllib.robotparser import RobotFileParser

import requests
from bs4 import BeautifulSoup, FeatureNotFound, XMLParsedAsHTMLWarning
from bs4.element import Tag

LOG = logging.getLogger(__name__)

DEFAULT_HTML_PARSER = "html.parser"
DEFAULT_CRAWL_DELAY = 3
USER_AGENT = f"babyspider/{importlib.metadata.version('babyspider')}"


def _log_exception(
    message: str,
    logger: Callable[[object], None],
    exception: type[BaseException],
    log_prefix: str = "",
) -> BaseException:
    """Log the prefix and message with the given logger then return a new instance of the given
    execption with the message from the given source execption, if given.

    Args:
        message: The base message to be logged and used to construct the execption
        logger: The log function that will be called with the combination of log prefix and message
            seprated by a space
        exception: The exception class that will be instanced with the message and returned

    Returns:
        An instance of the exception class called with message as the argument
    """
    logger(f"{log_prefix} {message}")
    return exception(message)


class Crawler:
    """A simple webcrawler that limits itself to a single domain.

    See ``crawl()`` to begin crawling a site.

    Attributes:
        base_output_dir: The base directory where the download directory will be created.
            Defaults to :class:``pathlib.Path()``

        parser: A valid BeautifulSoup HTML parser, typically one of `html.parser`, `lxml`,
            or 'html5lib'. Defaults to DEFAULT_HTML_PARSER

        crawl_delay: The number of seconds to pause between download requests. Defaults to
            DEFAULT_CRAWL_DELAY. Minimum value of 0

        use_sitemaps:  When sitemaps should be used, one of 'first', 'never', or 'only'.
            Defaults to 'first'

        max_redirects: The maximum number of redirects to follow when resolving URLs before
            downloading. Defaults to 10. Minimum 5 recommned by RFC 1945

        http_timeout: Amount of seconds to wait before giving up on any requests. Defaults to 300

        chunk_size: Size of chunks used when streaming downloads. Defaults to 8192

        index_filename: The name assigned when downloading directory indexes. Defaults to
            'index.html'

    """

    max_redirects = 10  # This should be at least 5 as per RFC 1945
    http_timeout = 300
    chunk_size = 8192

    # Used as the file name for directory indexes. While "index.html" is a pretty safe standard,
    # it's technically possible that therey might be non-index files called this and we'd end up
    # clobbering them, if the website both called its index files "default.html" and ALSO had
    # another file called "index.html". If they only have a different index name, we may end up
    # making a second copy with this name, but that's non-destructive, so it's not a big risk.
    index_filename = "index.html"

    def __init__(
        self,
        base_output_dir: str | Path = Path(),
        parser: str = DEFAULT_HTML_PARSER,
        crawl_delay: int = DEFAULT_CRAWL_DELAY,
        use_sitemaps: str = "first",
    ) -> None:
        """Initializes a new Crawler class and allows setting the base output directory, parser,
        crawl delay, and use of sitemaps.

        See the class docstring for details on these and all othe attributes.
        """
        self.check_parser(parser)
        self.parser = parser
        self.crawl_delay = max(int(crawl_delay), 0)
        self.base_output_dir = Path(base_output_dir)

        if use_sitemaps not in ("first", "only", "never"):
            LOG.warning(
                "The attribute use_sitemaps should only be set to 'first', 'never', or 'only', "
                "not %r. Setting it the default value 'first' instead.",
                use_sitemaps,
            )
            self.use_sitemaps = "first"
        else:
            self.use_sitemaps = use_sitemaps

        # These are state used during a crawl, but we set them here to make sure the attributes
        # exist even before a crawl.
        self.__known_urls: set[str] = set()
        self.__robot_parser = RobotFileParser()
        self.__current_crawl_delay = self.crawl_delay

    @staticmethod
    def check_parser(parser: str) -> None:
        """Return a ValueError if the given parser is not available in BeautifulSoup.

        Args:
            parser: The parser string to validate

        Raises:
            ValueError: If the requested `parser` is not available
        """
        try:
            BeautifulSoup("", parser)
        except FeatureNotFound as err:
            raise _log_exception(
                f"BeautifulSoup could not find the requested HTML parser {parser!r}: {err}",
                LOG.error,
                ValueError,
            ) from err

    @staticmethod
    def strip_url(url: str) -> str:
        """Return a URL without any params, query, or fragment components.

        Args:
            url: The URL to strip

        Raises:
            ValueError: Typically only raise when a non-string is given or
                :func:`urllib.parse.urlparse()` otherwise fails

        Returns:
            The URL stripped of params, query, and fragment, if any
        """
        try:
            parsed_url = urlparse(url)
        except (URLError, AttributeError, TypeError) as err:
            raise _log_exception(
                f"Could not parse URL to strip {url!r}: {err}",
                LOG.error,
                ValueError,
            ) from err
        return urlunparse([parsed_url.scheme, parsed_url.netloc, parsed_url.path, "", "", ""])

    @staticmethod
    def domain_from_url(url: str) -> str | None:
        """Return just the domain from a URL or None if there is no domain

        Without a scheme, the URL is just considered a path and no domain will be found!

        Args:
            url: The URL to extract a domain from

        Raises:
            ValueError: Typically only raise when a non-string is given or
                :func:`urllib.parse.urlparse()` otherwise fails

        Returns:
            The URL's domain, stripped of anhy username, password, and port
        """
        try:
            domain = urlparse(url).netloc
        except (URLError, AttributeError, TypeError) as err:
            raise _log_exception(
                f"Could not parse URL to extract domain from {url!r}: {err}",
                LOG.error,
                ValueError,
            ) from err

        # Remove any user and password.
        if "@" in domain:
            domain = domain[domain.index("@") + 1 :]
        # Remove any port number.
        if ":" in domain:
            domain = domain[: domain.index(":")]

        return domain if domain else None

    def get_canonical_url(self, url: str) -> str:
        """Get the final URL location following any redirects.

        This requires that the internal robots.txt file be configured. See
        :method`babyspider.crawler.Crawler.read_robots_txt()`

        Args:
            url: The URL to resolve

        Raises:
            requests.RequestException: Raise if any URL is forbidden by the robots.txt or
                if there are more than :attr`self.max_redirects` redirects

        Returns:
            The canonicalized URL after following any redirects
        """
        LOG.debug("Looking up canonical URL for: %r", url)

        # Since, in theory, there shouldn't be tons of redirects, we deliberately don't include the
        # wait time for HEAD requests, just when we actually download stuff with GET later.

        # Check that we are allowed to hit this URL.
        if not self.__robot_parser.can_fetch(USER_AGENT, url):
            raise _log_exception(
                f"Request URL {url!r} is forbidden by robots.txt.",
                LOG.warning,
                requests.RequestException,
            )

        response = requests.head(url, headers={"User-Agent": USER_AGENT}, timeout=self.http_timeout)

        # Handle any redirecst.
        redirects = 0
        new_url = url
        while redirects < self.max_redirects and response.is_redirect:
            redirects += 1

            old_url = new_url
            # The code for response.is_redirect includes checking that location is in the headers.
            new_url = response.headers["Location"]

            # Ensure the new location is a full URL.
            if "://" not in new_url:
                parsed_old_url = urlparse(old_url)

                # Turn a relative path to an absolute one.
                if not new_url.startswith("/"):
                    new_url = str(Path(parsed_old_url.path).parent / new_url)

                new_url = urlunparse(
                    [parsed_old_url.scheme, parsed_old_url.netloc, new_url, "", "", ""]
                )

            # Check that we are allowed to hit this URL.
            if not self.__robot_parser.can_fetch(USER_AGENT, url):
                raise _log_exception(
                    f"Request URL {url!r} redirects to URL {new_url!r} that "
                    f"is forbidden by robots.txt.",
                    LOG.warning,
                    requests.RequestException,
                )
            LOG.debug("Following redirect to: %r", new_url)
            response = requests.head(
                new_url,
                headers={"User-Agent": USER_AGENT},
                timeout=self.http_timeout,
            )

        if response.is_redirect:
            raise _log_exception(
                (
                    f"Got more than {self.max_redirects} redirects trying to resolve url {url!r}. "
                    f"Last redirect URL: {response.headers['Location']!r}"
                ),
                LOG.error,
                requests.RequestException,
            )

        # Ensure the final location is a full URL.
        new_url = response.url
        if "://" not in new_url:
            parsed_old_url = urlparse(old_url)

            # Turn a relative path to an absolute one.
            if not new_url.startswith("/"):
                new_url = str(Path(parsed_old_url.path).parent / new_url)

            new_url = urlunparse(
                [parsed_old_url.scheme, parsed_old_url.netloc, new_url, "", "", ""]
            )

        # Only report if the URL actually changed.
        if new_url != url:
            LOG.info("Canonical URL: %r", new_url)
        return new_url

    def get_local_path(self, url: str) -> Path:
        """Determine a relative local path and file name for a given URL.

        Args:
            url: The URL to determine a local file path from

        Raises:
            ValueError: Typically only raise when a non-string is given or
                :func:`urllib.parse.urlparse()` otherwise fails

        Returns:
            A relative local path based on the URL with the default `self.index_filename`
            added if the path from the URL is a directory.filename.
        """
        try:
            parsed_url = urlparse(url)
        except (URLError, AttributeError, TypeError) as err:
            raise _log_exception(
                f"Could not parse path from canonical URL {url!r}: {err}",
                LOG.error,
                ValueError,
                "Skipping download.",
            ) from err

        # We explicitly are ignoring any params or query (as well as fragment), so the path is
        # enough to determine the file location.
        # But, the path component of a URL always starts with a leading slash, which we need to
        # strip or our local Path object will be absolute and won't let us prepend the output
        # directory.
        local_path = Path(parsed_url.path.lstrip("/"))

        # A Path object will strip the trailing slash that indicates this is a directory resource,
        # so if we are pulling the index of a directory, add an index file name to the local path.
        if parsed_url.path.endswith("/"):
            local_path = local_path / self.index_filename

        return local_path

    def download_resource(self, url: str, output_dir: str | Path) -> tuple[Path, str]:
        """Download the URL to the appropriate path in the output directory, creating
        subdirectories as needed. Return the path to the created file its canoncial
        source URL. Skip downloading if the file already exists.


        Args:
            url: The URL of the resources to download
            output_dir: The output directory to start at when determining where to place the
                resource locally. The full path will include any subdirectories from the URL path.
                For example, if `url` is `http://example.com/some/file.html` and the `output_dir`
                is `/tmp`, the file will be saved as `/tmp/some/file.html`.

        Raises:
            IOError: If the URL could not be normalized by ``get_canonical_url()`` or there
                is a problem creating intermidiate directories or writing the file to disk
            ValueError: If the URL cannot be stripped, is on a different domain after redirects,
                or already exists locally but as a directory instead of a file
            RequestException: If there is a network issue downloading the file

        Returns:
            A tuple of the path where the file was downloaded and the final URL it was pulled from
        """
        LOG.info("Downloading URL: %r", url)

        # First, find the canoncial URL so we can determine the check if the file is downloaded
        # already.
        try:
            final_url = self.strip_url(self.get_canonical_url(url))
        except (ValueError, requests.RequestException) as err:
            LOG.error(
                "Skipping download. Could not get stripped, canonical URL for %r: %s",
                url,
                err,
            )
            raise

        # Make sure the final URL hasn't gone to some other domain than where we started.
        if self.domain_from_url(url) != self.domain_from_url(final_url):
            raise _log_exception(
                f"Canonical URL for {url!r} is {final_url!r} which is on a different domain.",
                LOG.info,
                ValueError,
                "Skipping download.",
            )

        local_path = Path(output_dir) / self.get_local_path(final_url)

        # Skip downloading the file if it exists.
        # This is both to prevent downloading the same thing from multiple links, and also to allow
        # us to continue partially succeeded attempts without redownloading everything.
        if local_path.exists():
            if not local_path.is_file():
                raise _log_exception(
                    (
                        f"Resource {url!r} from final URL {final_url!r} already exists as "
                        f"{local_path!r}, but not as a file!"
                    ),
                    LOG.error,
                    ValueError,
                    log_prefix="Skipping download.",
                )
            LOG.info("Skipping download. Resource already exists as: %r", str(local_path))
            return local_path, final_url

        # Download the resource if it doesn't exist.
        # First make sure we have any needed subdirectories created.
        LOG.debug("Creating local directories if needed: %r", str(local_path.parent))
        local_path.parent.mkdir(parents=True, exist_ok=True)

        # Then wait a bit just before we actually perform the download to prevent us from
        # spamming the server with requests.
        if self.__current_crawl_delay:
            LOG.debug("Delaying %s seconds.", self.__current_crawl_delay)
            time.sleep(self.__current_crawl_delay)

        # We don't have to check the robots.txt because pulling the canonical URL will have already
        # checked for us.
        LOG.info("Downloading resource from %r to: %r", final_url, str(local_path))
        with requests.get(
            url,
            stream=True,
            headers={"User-Agent": USER_AGENT},
            timeout=self.http_timeout,
        ) as response:
            try:
                response.raise_for_status()
            except requests.RequestException as err:
                LOG.error("Could not download resource at URL %r: %s", url, err)
                raise
            with open(local_path, "wb") as file_handle:
                for chunk in response.iter_content(chunk_size=self.chunk_size):
                    file_handle.write(chunk)
        return local_path, final_url

    def get_new_domain_urls(
        self,
        file_path: Path,
        base_url: str,
        images: bool = False,
        known_urls: set[str] | None = None,
    ) -> set[str]:
        """Scan the given file and, if it is parsable as HTML, extract and canonicalize any URLs to
        other resources for the base URLs domain not already known.

        Args:
            file_path: The path to the file to scan for URLs
            base_url: The original URL the file was downloaded from. Used to convert relative links
                in the file into complete URLs and to determine if absolute URLs are on the same
                domain and can be included
            images: If `True`, Also extract URLs for image tags
            known_urls: A set of URLs to exclude from the returned result

        Raises:
            IOError: If there are problems reading the file

        Returns:
            The set of new URLs found in the file. This will be empty if the file could not be
            parsed as HTML
        """

        LOG.info("Scanning downloaded resource: %r", str(file_path))
        try:
            # BeautifulSoup will automatically attempt to determine the encoding, so we don't
            # have to specify it here.
            with open(file_path) as file_handle:  # pylint: disable=unspecified-encoding
                soup = BeautifulSoup(file_handle, self.parser)
        except IOError as err:
            LOG.error(
                "Skipping URL detection. Could not read %r: %s",
                file_path,
                err,
            )
            raise
        except (AttributeError, IndexError, UnicodeDecodeError) as err:
            LOG.info(
                "Skipping URL detection. Could not parse %r as HTML.",
                str(file_path),
            )
            LOG.debug(
                "Could not parse %r as HTML using parser %r: %s",
                str(file_path),
                self.parser,
                err,
            )
            return set()

        # We specifically ask for the known URLs rather than pull from `self.__known_urls` here
        # so that this method is usable on its own, not just internally.
        if not known_urls:
            known_urls = set()

        base_domain = self.domain_from_url(base_url)
        new_domain_urls = set()

        tag_types = {"a": "href"}
        if images:
            tag_types["img"] = "src"

        for tag_type, tag_attr in tag_types.items():
            for tag in soup.find_all(tag_type, {tag_attr: True}):
                try:
                    if not isinstance(tag, Tag):
                        continue
                    target_url = str(tag.get(tag_attr))

                    # If this is a relative URL, resolve it to its absolute path.
                    if not urlparse(target_url).netloc:
                        target_url = urljoin(base_url, target_url)

                    # Check the domain of the resolved URL and add it to our list if it matches.
                    if self.domain_from_url(target_url) == base_domain:
                        # Explicitly strip any params, query, and fragment components.
                        target_url = self.strip_url(target_url)
                        if target_url not in known_urls:
                            LOG.debug(
                                "Found new URL on domain %r: %r",
                                base_domain,
                                target_url,
                            )
                            new_domain_urls.add(target_url)
                except (URLError, AttributeError, TypeError) as err:
                    LOG.warning(
                        "Skipping URL domain check. Could not parse URL from %r: %s",
                        tag,
                        err,
                    )

        LOG.info("Found %s new URLs on the same domain.", len(new_domain_urls))
        return new_domain_urls

    def __crawl(self, url: str, output_dir: Path, images: bool = False, errors: int = 0) -> int:
        """Download the resource at the given URL and, if it is HTML, do the same for all linked
        resources on the URL's same domain. Returns the number of errors.

        This is meant to be called only by the wrapper method `self.crawl()` that validates the
        determines the starting URL, resets internal found URL tracking, sets up the robots.txt
        parser before starting the crawl.


        Args:
            url: The URL to download and scan
            output_dir: The base path where the resource will be downloaded. Passed to
                ``download_resource()``
            images: If `True`, scan image tags for URLs, not just anchor tags
            errrors: Used as the starting number of errors to be added to before returning

        Raises:
            IOError: If there is a problem saving the resource to disk or reading from it
                afterwards
            ReqeustException: If there is a problem dowloading the resource

        Returns:
            The number of errors passed in plus any additional errors have have been tallied
            while processing this URL
        """
        try:
            file_path, canonical_url = self.download_resource(url, output_dir)
        except (
            requests.RequestException,
            IOError,
        ) as err:
            # We should have already dispayed the problem as an error level log message.
            # this is just a backup incase something went really wrong.
            LOG.debug("Download of URL %r failed: %s", url, err)
            return errors + 1
        except ValueError:
            # If the URL just lead to a different domain, it's not really an error.
            return errors
        self.__known_urls.add(canonical_url)

        try:
            found_urls = self.get_new_domain_urls(
                file_path, canonical_url, images, self.__known_urls
            )
            self.__known_urls.update(found_urls)
            for found_url in found_urls:
                errors = self.__crawl(found_url, output_dir, images, errors)
        except IOError as err:
            LOG.error(
                "IOError while crawling URLs from %r in %r: %s",
                canonical_url,
                file_path,
                err,
            )
            return errors + 1

        return errors

    def _get_starting_and_robots_urls(self, url_or_domain: str) -> tuple[str, str]:
        """Convert the given URL or domain into a valid, canonical URL, and return it and the
        standad robot.txt URL for its domain.


        Args:
            url_or_domain: The URL or domain name to validate

        Raises:
            ValueError: If the URL is not parsable or has no domain

        Returns:
            A tuple of the validated starting URL and the robots.txt URL
        """

        try:
            parsed_url = urlparse(url_or_domain)
        except (URLError, AttributeError, TypeError) as err:
            raise _log_exception(
                f"Could not parse {url_or_domain!r} as a URL or plain string!",
                LOG.error,
                ValueError,
            ) from err

        # If it has a net location, it's should also a scheme and is a URL we can use
        if parsed_url.netloc:
            return (
                self.strip_url(url_or_domain),
                urlunparse([parsed_url.scheme, parsed_url.netloc, "/robots.txt", "", "", ""]),
            )

        # If it had a scheme but no net location (for example: `file:///etc`), we can't
        # use it because we won't know the domain to restrict crawling to
        if parsed_url.scheme:
            raise _log_exception(
                f"Domain seemes missing from starting URL: {url_or_domain!r}",
                LOG.error,
                ValueError,
            )

        # But if there's not net location or scheme, it's probably just a domain name, so
        # stick `https://` in front and try again.
        LOG.info(
            "No net location found in %r, assuming it's just a domain name and adding HTTPS.",
            url_or_domain,
        )
        url_or_domain = f"https://{url_or_domain}"
        try:
            parsed_url = urlparse(url_or_domain)
        except (URLError, AttributeError, TypeError) as err:
            # I have no idea how we could get here, but better safe than sorry.
            raise _log_exception(
                f"Could not parse {url_or_domain!r} as a URL or plain string!",
                LOG.error,
                ValueError,
            ) from err

        if parsed_url.netloc:
            return (
                self.strip_url(url_or_domain),
                urlunparse([parsed_url.scheme, parsed_url.netloc, "/robots.txt", "", "", ""]),
            )

        # If it still has no net location, give up.
        # Again, this part seems highly unlikely, but it's safer than not knowing it happened.
        raise _log_exception(
            "Domain seemes missing from starting URL even after adding HTTPS scheme: "
            f"{url_or_domain!r}",
            LOG.error,
            ValueError,
        )

    def _make_output_dir(self, output_dir_name: str | Path, add_timestamp: bool | None) -> Path:
        """Determine and create, if needed, the output download directory under the base directory,
        returning a Path object of the full output path.

        Args:
            output_dir_name: The directory to be created under the base output directory
            add_timestamp: If `True`, add a timestamp to the end of the directory name

        Raises:
            TypeError: If the output directory name is the wrong type
            ValueError: If the directory to be created already exists as a file
            IOError: If something goes wrong with the directory creation itself

        Returns:
            The full path of the created directory
        """
        try:
            path = Path(self.base_output_dir) / Path(output_dir_name)
            if add_timestamp:
                path = Path(str(path) + f"_{datetime.now():%Y%m%d_%H%M}")
        except TypeError as err:
            raise _log_exception(
                (
                    "Cannot build output directory name from output path "
                    f"{self.base_output_dir!r} and directory {output_dir_name!r}: {err!r}"
                ),
                LOG.error,
                TypeError,
            ) from err

        if path.exists():
            if path.is_dir():
                LOG.info("Using existing output directory: %r", str(path))
                return path
            raise _log_exception(
                (f"Output directory path {str(path)!r} exists, but is not a directory!"),
                LOG.error,
                ValueError,
            )

        LOG.info("Creating domain output directory: %r", str(path))
        path.mkdir(parents=True)
        return path

    def __get_minimum_crawl_delay(self) -> int:
        """Determine the minimum crawl delay based on the requested crawl delay and any
        requests per second or crawl delay from the robots.txt file.

        Returns:
            The minimum permitted crawl delay in seconds
        """
        # Include a zero second delay in case we get any sort of negative number nonsense.
        delays = [self.crawl_delay]
        LOG.info("Requested crawl delay is: %s", self.crawl_delay)

        # Convert the requests per second, if we got one, into a minmum delay between requests.
        if robot_rps := self.__robot_parser.request_rate(USER_AGENT):
            try:
                delays.append(math.ceil(robot_rps[1] / robot_rps[0]))
                LOG.info(
                    "Found robots.txt max requests per second (as minimum crawl delay) of: %s",
                    delays[-1],
                )
            except (TypeError, ZeroDivisionError) as err:
                LOG.warning(
                    "Couldn't determine crawl delay from requests per second %r: %s",
                    robot_rps,
                    err,
                )
        if robot_crawl_delay := self.__robot_parser.crawl_delay(USER_AGENT):
            try:
                delays.append(math.ceil(int(robot_crawl_delay)))
                LOG.info("Found robots.txt minimum crawl delay of: %s", delays[-1])
            except ValueError as err:
                LOG.warning(
                    "Couldn't parse and round up crawl delay %r: %s",
                    robot_crawl_delay,
                    err,
                )

        minimum_delay = max(delays)

        # Only bother to report on the actual crawl delay if we got more than the requested one.
        if len(delays) > 1:
            LOG.info("Using crawl delay of: %s", minimum_delay)

        return minimum_delay

    def get_urls_from_sitemap(self, sitemap: bytes) -> set[str]:
        """Return all the URLs from a site map in a set.

        Args:
            sitemap: The raw bytes of the contents of a sitemap to extract URLs from.

        Returns:
            The set of URLs extracted
        """
        try:
            soup = BeautifulSoup(sitemap, "xml")
        except (AttributeError, IndexError, UnicodeDecodeError) as err:
            LOG.error("Skipping sitemap URL parsing. Could not parse sitemap XML: %s", err)
            return set()
        except FeatureNotFound:
            try:
                LOG.info("No XML parser found. Parsing sitemap with HTML parser.")
                # Technically sitemaps are XML, but it's pretty simple and all we're doing is
                # grabbing the URLs from `loc` tags. Because of that, we'll use an HTML parser
                # and just diable the warning for the moment.
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=XMLParsedAsHTMLWarning)
                    soup = BeautifulSoup(sitemap, self.parser)
            except (AttributeError, IndexError, UnicodeDecodeError) as err:
                LOG.error(
                    "Skipping sitemap URL parsing. No XML parser and could not parse sitemap as "
                    "HTML with parser %r: %s",
                    self.parser,
                    err,
                )
                return set()

        return {tag.get_text() for tag in soup.find_all("loc")}

    def download_sitemap_urls(self, sitemap_url: str, output_dir: str | Path) -> int:
        """Download all the URLs from the given site map URL.

        Args:
            sitemap_url: The URL of the sitemap to download, read, and then download the URLs in it
            output_dir: The parent directory where downloaded sitemap URLs will be saved. Passed
                to ``download_resource()``

        Returns:
            The number of errors that occurred. This may be one for every URL in the sitemap if
            they all failed, or literlly only one if downloading the sitemap itself failed
        """

        LOG.info("Downloading sitemap: %r", sitemap_url)
        # We assume that any given sitemap is allowed for download by the robots.txt rules.
        try:
            response = requests.get(sitemap_url, timeout=self.http_timeout)
        except requests.RequestException as err:
            LOG.error("Download of sitemap %r failed: %s", sitemap_url, err)
            return 1

        LOG.info("Downloading URLs listed in sitemap.")
        errors = 0
        for url in self.get_urls_from_sitemap(response.content):
            self.__known_urls.add(url)
            try:
                # We don't need the file path since we are just downloading everything from the
                # sitemap and not scanning anything for more links.
                canonical_url = self.download_resource(url, output_dir)[1]
                self.__known_urls.add(canonical_url)
            except (requests.RequestException, IOError) as err:
                # We should have already dispayed the problem as an error level log message.
                # this is just a backup incase something went really wrong.
                LOG.debug("Download of URL %r from sitemap failed: %s", url, err)
                errors += 1
            except ValueError as err:
                # If the URL just lead to a different domain, it's not really an error.
                pass
        return errors

    def read_robots_txt(self, robots_url: str) -> None:
        """Read the robots.txt file into a new internal parser for use.

        This is required before using other methods such as get_canonical_url().

        Args:
            robots_url: The URL of the robots.txt file to read

        Raises:
            URLError: Usually only if something goes wrong with an SSL handshake while downloading
        """

        LOG.info("Reading robots.txt file, if it exists, from: %r", robots_url)
        self.__robot_parser = RobotFileParser(robots_url)
        try:
            self.__robot_parser.read()
        except URLError as err:
            raise _log_exception(
                f"URL error attempting to read robots.txt from {robots_url!r}: {err}",
                LOG.error,
                ValueError,
                "Maybe try plain HTTP?",
            ) from err

    def crawl(
        self,
        url_or_domain: str,
        output_dir_name: str | None = None,
        images: bool = False,
        add_timestamp: bool | None = None,
    ) -> tuple[str, str, int]:
        """Download the resource at the given URL and, if it is HTML, do the same for all linked
        resources on the same domain. Returns the full output path, the canonicalized starting URL,
        and the number of errors.

        Args:
            url_or_domain: The URL or domain to start the web crawl from
            output_dir_name: The output directory to create under  base output director
                where found resources will be downloaded to. Defaults to the domain name of the
                starting URL
            images: If `True`, scan image tags for URLs to download, not just links
            add_timestamp: If `True`, add a timestamp to the end of the output directory name.
                This defaults to `False` except when no output directory name is given

        Raises:
            ValueError: If the starting URL is forbidden by the site's robots.txt file, has too
            many redirects, or if `use_sitemaps` is set to `only` but the are no sitemaps

        Returns:
            The name of the full output directory, the canonical starting URL, and the total number
            of errors during the process

        This is a wrapper function around `self.__crawl()` that validates the determines the
        starting URL, resets internal found URL tracking, sets up the robots.txt parser before
        starting the crawl.
        """

        # Warn if we are asked to crawl for images but also to only use sitemaps.
        if images and (self.use_sitemaps == "only"):
            LOG.warning(
                "Crawling for images is set, but we're not crawling for URLs at all, only "
                "downloading from sitemaps! Image crawlings only happens when sitemaps is set "
                "to 'first' or 'never'."
            )

        starting_url, robots_url = self._get_starting_and_robots_urls(url_or_domain)

        # Get the robots.txt file, if any.
        self.read_robots_txt(robots_url)

        # See if the starting URL is one we can request.
        if not self.__robot_parser.can_fetch(USER_AGENT, starting_url):
            raise _log_exception(
                f"Starting URL {starting_url!r} is forbidden by robots.txt file!",
                LOG.error,
                ValueError,
                "Maybe try the root URL?",
            )

        # Get the canonical version of the starting URL after any redirects.
        try:
            starting_url = self.strip_url(self.get_canonical_url(starting_url))
        except requests.RequestException as err:
            raise _log_exception(
                f"Error determing canonical location of starting URL {url_or_domain!r}: {err}",
                LOG.error,
                ValueError,
            ) from err
        LOG.info("Canonical starting URL: %r", starting_url)

        # Make sure we have an output directory name, using the domain name of the starting URL
        # as a default.
        if not output_dir_name:
            output_dir_name = self.domain_from_url(starting_url)
            if add_timestamp is None:
                add_timestamp = True
        # For now, mypy can't detect that output_dir_name is not None without this assert.
        assert output_dir_name is not None
        output_path = self._make_output_dir(output_dir_name, add_timestamp)

        # Set a minimum crawl delay based on the configured one and the robots.txt settings.
        self.__current_crawl_delay = self.__get_minimum_crawl_delay()

        # Seed the known URLs with our starting URL and begin crawling.
        errors = 0
        self.__known_urls = set()

        # If there is a sitemap, and we've been set to use it, start with that.
        if self.use_sitemaps != "never":
            if sitemap_urls := self.__robot_parser.site_maps():
                LOG.info("Found %s robots.txt sitemaps URLs.", len(sitemap_urls))
                LOG.debug("Sitemaps URLs: %r", sitemap_urls)
                for sitemap_url in sitemap_urls:
                    errors += self.download_sitemap_urls(sitemap_url, output_path)
            else:
                LOG.info("No sitemaps found.")
                if self.use_sitemaps == "only":
                    raise _log_exception(
                        "Requested downloading sitemap URLs only, but there are no sitemaps!",
                        LOG.error,
                        ValueError,
                    )
        LOG.debug("Errors after sitemap URL downloads: %s", errors)

        # Crawl the URL unless we're doing the sitemaps only.
        if self.use_sitemaps != "only":
            self.__known_urls.add(starting_url)
            LOG.info(
                "Downloading links%s from %r and saving into: %r",
                " and images" if images else "",
                starting_url,
                str(output_path),
            )
            errors += self.__crawl(starting_url, output_path, images, errors)

        return str(output_path), starting_url, errors

    def __str__(self) -> str:
        """Return a pretty string of the class.

        Returns:
            Pretty string name of the instance
        """
        return f"{self.__class__.__name__}({str(self.base_output_dir)!r})"

    def __repr__(self) -> str:
        """Return a represenational string of the class.

        Returns:
            Python representational string of the instance
        """
        return (
            f"<{self.__class__.__name__}(base_output_dir={self.base_output_dir!r}, "
            f"parser={self.parser!r}, crawl_delay={self.crawl_delay}, "
            f"use_sitemaps={self.use_sitemaps!r})>"
        )
