"""API service wrapper for babyspider"""

import subprocess
from concurrent.futures import ThreadPoolExecutor, Future
from functools import wraps
from typing import Callable
from fastapi import FastAPI, HTTPException


app = FastAPI()

# Keyed to the URL or domain reqeusted. Should contain a status and payload.
REQUESTS: dict[str, dict[str, str | dict[str, str | None] | Future[object] | None]] = {}

_THREAD_POOL_EXECUTOR = ThreadPoolExecutor()


def futurepool(function: Callable[..., object]) -> object:
    """A decorator to wrap a function in a future."""

    @wraps(function)
    def wrap(*args: object, **kwargs: object) -> Future[object]:
        """A wrapper to run the function in a future."""
        return (_THREAD_POOL_EXECUTOR).submit(function, *args, **kwargs)

    return wrap


@futurepool
def run_crawl(url_or_domain: str) -> str:
    """Run the crawler in a future and capture the output directory."""
    return subprocess.run(
        ["poetry", "run", "babyspider", "-qq", url_or_domain], check=False, stdout=subprocess.PIPE
    ).stdout.decode("utf-8")


def get_request_state(url_or_domain: str) -> tuple[str, object | None]:
    """Return the current status and payload of a request."""
    if not REQUESTS[url_or_domain]["future"].done():  # type: ignore
        return "running", None
    return "done", REQUESTS[url_or_domain]["future"].result()  # type: ignore


@app.get("/")
async def root() -> dict[str, str | dict[str, object] | None]:
    """Return the current status of all known requests."""
    for url_or_domain, request in REQUESTS.items():
        request["status"], request["payload"] = get_request_state(url_or_domain)  # type: ignore
    return {
        "message": "Current known requests",
        "payload": {
            url_or_domain: {
                "status": request["status"],
                "payload": request["payload"],
            }
            for url_or_domain, request in REQUESTS.items()
        },
    }


@app.get("/crawl")
async def crawl(url_or_domain: str) -> dict[str, str]:
    """Take a url or domain, run babyspider and return a request ID."""

    if url_or_domain in REQUESTS:
        if REQUESTS[url_or_domain]["status"] == "running":
            raise HTTPException(400, f"Crawl reqeuest for {url_or_domain} already running!")
        # We did this before.
        # TODO: Determine if we want to notify the end user of that.

    REQUESTS[url_or_domain] = {"status": "running", "payload": None, "future": None}

    # Start the crawl in a future.
    REQUESTS[url_or_domain]["future"] = run_crawl(url_or_domain)  # type: ignore

    return {
        "message": f"Getting {url_or_domain}",
    }


@app.get("/results")
async def results(url_or_domain: str) -> dict[str, str | None]:
    """Return the results of the requested crawl."""
    # Check if the request ID is valid.
    if url_or_domain not in REQUESTS:
        raise HTTPException(404, f"Unknown request ID {url_or_domain}!")
    REQUESTS[url_or_domain]["status"], REQUESTS[url_or_domain]["payload"] = get_request_state(
        url_or_domain
    )  # type: ignore
    return {
        "message": f"Got {url_or_domain}",
        "status": REQUESTS[url_or_domain]["status"],  # type: ignore
        "payload": REQUESTS[url_or_domain]["payload"],  # type: ignore
    }
