"""babyspider - A single-domain web crawler example project"""

import argparse
import logging
import importlib
import sys
from datetime import datetime
from pathlib import Path
from typing import Sequence
from urllib.parse import urlparse

import loggery

from babyspider.crawler import Crawler, DEFAULT_HTML_PARSER, DEFAULT_CRAWL_DELAY

__version__ = importlib.metadata.version("babyspider")
__author__ = "erskin@eldritch.org"

LOG = logging.getLogger(__name__)


def build_parser() -> argparse.ArgumentParser:
    """Build and return a command line argument parser for the babyspider webcrawler.

    Returns:
        The configured babyspider argument parser
    """
    parser = argparse.ArgumentParser(prog="babyspider", description="A single-domain web crawler")

    parser.add_argument(
        "url",
        help="A domain name or URL to start crawling from. SSL will be used for domain names",
    )
    parser.add_argument(
        "-o",
        "--output-path",
        help="The base path the output directory will be created in. "
        "Defaults to the current directory",
        default="",
    )
    parser.add_argument(
        "-p",
        "--parser",
        help=f"The BeautifulSoup HTML parser to use. Defaults to {DEFAULT_HTML_PARSER!r}",
        default=DEFAULT_HTML_PARSER,
    )
    parser.add_argument(
        "--fast",
        help="Alias for `--parser lxml`",
        action="store_const",
        const="lxml",
        dest="parser",
    )
    parser.add_argument(
        "--good",
        help="Alias for `--parser html5lib`",
        action="store_const",
        const="html5lib",
        dest="parser",
    )
    parser.add_argument(
        "-i",
        "--images",
        help="Download image URLs, not just links",
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "-I",
        help="A short option verison of --no-images",
        action="store_false",
        dest="images",
    )
    parser.add_argument(
        "-s",
        "--use-sitemaps",
        help="When to download from any sitemaps. 'only' means no crawling. Defaults to 'first'",
        choices=("first", "only", "never"),
        default="first",
    )
    parser.add_argument(
        "-D",
        "--crawl-delay",
        help=f"Number of seconds to wait between requests. Defaults to {DEFAULT_CRAWL_DELAY}",
        default=DEFAULT_CRAWL_DELAY,
        type=int,
    )
    parser.add_argument(
        "-d",
        "--directory-name",
        help="The name of the output directory. Defaults to the domain name with a timestamp",
    )
    parser.add_argument(
        "-t",
        "--timestamp",
        help="Append the date and time to end of the output directory name. This is the "
        "default behavior if no directory name is given",
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "-T",
        help="A short option verison of --no-timestamp",
        action="store_false",
        dest="timestamp",
    )
    loggery.add_logging_args(parser)
    parser.add_argument("--version", action="version", version="%(prog)s " + __version__)

    return parser


def run(args: Sequence[str] | None = None) -> int:
    """Run the babyspider web crawler as a command line tool with the given arguments.

    Args:
        args: An sequence of command line argument strings. This defaults
            to ``sysy.argv``

    Returns:
        Either 0 if no errors or 1 if errors occurred
    """

    parsed_args = build_parser().parse_args(args)
    loggery.configure_logging_from_args(
        parsed_args.verbose, parsed_args.quiet, parsed_args.log_stdout
    )
    LOG.debug("Parsed arguments: %r", parsed_args)

    # Crawl the URL and download the files.
    try:
        crawler = Crawler(
            parsed_args.output_path,
            parsed_args.parser,
            parsed_args.crawl_delay,
            parsed_args.use_sitemaps,
        )
        output_path, starting_url, errors = crawler.crawl(
            parsed_args.url,
            parsed_args.directory_name,
            parsed_args.images,
            parsed_args.timestamp,
        )
    except (TypeError, ValueError) as err:
        # We should have already displayed any relevant error message before we get here.
        LOG.debug(
            "Swallowing known exception to hide stack trace in command line output: %s",
            err,
        )
        # Since we're running as a command line tool, return 1 as an exit code to say something
        # went wrong.
        return 1

    if errors:
        LOG.error(
            "Download of web site resources from %r to %r completed with %s errors.",
            starting_url,
            output_path,
            errors,
        )
        # Since we're running as a command line tool, return 1 as an exit code to say something
        # went wrong.
        print(output_path)
        return 1

    # Since we're running as a command line tool, return 0 as an exit code to say things went fine.
    LOG.info(
        "Download of web site resources from %r to %r completed.",
        starting_url,
        output_path,
    )
    print(output_path)
    return 0
