"""test_babyspider.py - Test the babyspider module at the top level."""

import logging
from argparse import Namespace

from pytest import LogCaptureFixture
from pytest_mock import MockerFixture

import babyspider


def test_arg_parser() -> None:
    """Check that babyspider can build an argument parser and parse args."""

    # We only have a couple fixtures here because the goal is not to test Python's
    # argparse module, just make sure we don't break the expected set of command line
    # arguments.
    given_and_expected = (
        (
            ["-vviT", "--use-sitemaps", "only", "--crawl-delay", "5", "spork.org"],
            Namespace(
                url="spork.org",
                output_path="",
                parser="html.parser",
                images=True,
                use_sitemaps="only",
                crawl_delay=5,
                directory_name=None,
                timestamp=False,
                verbose=2,
                quiet=0,
                log_stdout=False,
            ),
        ),
        (
            [
                "-vt",
                "-s",
                "never",
                "--parser",
                "lxml",
                "-o",
                "/tmp",
                "http://www.eldritch.org",
            ],
            Namespace(
                url="http://www.eldritch.org",
                output_path="/tmp",
                parser="lxml",
                images=None,
                use_sitemaps="never",
                crawl_delay=3,
                directory_name=None,
                timestamp=True,
                verbose=1,
                quiet=0,
                log_stdout=False,
            ),
        ),
        (
            [
                "--log-stdout",
                "--no-images",
                "-vvqqq",
                "--quiet",
                "gopher://example.com/marble",
            ],
            Namespace(
                url="gopher://example.com/marble",
                output_path="",
                parser="html.parser",
                images=False,
                use_sitemaps="first",
                crawl_delay=3,
                directory_name=None,
                timestamp=None,
                verbose=2,
                quiet=4,
                log_stdout=True,
            ),
        ),
    )

    for given, expected in given_and_expected:
        results = babyspider.build_parser().parse_args(given)
        assert results == expected


def test_run(mocker: MockerFixture, caplog: LogCaptureFixture) -> None:
    """Test that we setup can call the crawler correctly."""

    mock_crawl = mocker.patch.object(babyspider.crawler.Crawler, "crawl")

    # Test the happy path first.
    mock_crawl.return_value = ("fake_path", "https://fake.example.com/", 0)
    result = babyspider.run(["-vd", "fake_path", "fake.example.com"])
    mock_crawl.assert_called()
    assert result == 0, "babyspider.run() should return 0 if the crawler returns 0 errors."
    assert (
        "from 'https://fake.example.com/' to 'fake_path' completed." in caplog.records[-1].message
        and caplog.records[-1].levelno == logging.INFO
    ), "Log output should include an info message with the source URL and output path on success."

    # Test the crawler returning with errors.
    caplog.clear()
    mock_crawl.reset()
    mock_crawl.return_value = ("fake_path", "https://fake.example.com/", 5)
    result = babyspider.run(["-vd", "fake_path", "fake.example.com"])
    mock_crawl.assert_called()
    assert (
        result == 1
    ), "babyspider.run() should return 1 if the crawler returns more than 0 errors."

    assert (
        "from 'https://fake.example.com/' to 'fake_path' completed with 5 errors."
        in caplog.records[-1].message
        and caplog.records[-1].levelno == logging.ERROR
    ), (
        "Log output should include an error message with the source URL, the output path, and the "
        "number of errors if non-zero."
    )
    # Test the craler throwing a ValueError exception.
    caplog.clear()
    mock_crawl.reset()
    mock_crawl.side_effect = ValueError("Fake ValueError for a bad URL")
    result = babyspider.run(["-vd", "fake_path", "fake.example.com"])
    mock_crawl.assert_called()
    assert result == 1, "babyspider.run() should return 1 if the crawler raises a ValueError."

    # Test the craler throwing a ValueError exception.
    caplog.clear()
    mock_crawl.reset()
    mock_crawl.side_effect = ValueError("Fake ValueError for a bad URL")
    result = babyspider.run(["-vd", "fake_path", "fake.example.com"])
    mock_crawl.assert_called()
    assert result == 1, "babyspider.run() should return 1 if the crawler raises a TypeError."
