"""test_crawler___crawl_method.py - Test the babyspider.crawler.Crawler.__crawl method."""

from pathlib import Path
from unittest.mock import call, ANY

from pytest_mock import MockerFixture

from babyspider.crawler import Crawler


def test___crawl_happy_path(mocker: MockerFixture) -> None:
    """Test the Crawler.__crawl() method when things go well."""

    # Mock the downloading and scanning for URLs to fake a simple website with an index and two
    # other pages. We simulate that the index has a link to page 1 but not page 2, # that page 1
    # links to both the index and page 2, while page 2 only links back to the index.
    mock_download = mocker.patch.object(Crawler, "download_resource")
    mock_download.side_effect = [
        (Path("fake_path/index.html"), "http://example.com/index.html"),
        (Path("fake_path/fake_page_1.html"), "http://example.com/fake_page_1.html"),
        (Path("fake_path/fake_page_2.html"), "http://example.com/fake_page_2.html"),
    ]
    mock_get_urls = mocker.patch.object(Crawler, "get_new_domain_urls")
    mock_get_urls.side_effect = [
        ("http://example.com/fake_page_1.html",),
        ("http://example.com/fake_page_2.html",),
        tuple(),
    ]

    crawler = Crawler()
    output_dir = Path("fake_path/")
    # pylint: disable-next=protected-access
    errors = crawler._Crawler__crawl("http://example.com/", output_dir)  # type: ignore

    assert errors == 0

    # We don't care how the code does it, just that it tries to download each file once.
    assert mock_download.call_count == 3
    mock_download.assert_has_calls(
        (
            call("http://example.com/", output_dir),
            call("http://example.com/fake_page_1.html", output_dir),
            call("http://example.com/fake_page_2.html", output_dir),
        ),
        any_order=True,
    )

    # Simiarly, we don't care how the code does it, just that it tries to scan each file once.
    assert mock_get_urls.call_count == 3
    mock_get_urls.assert_has_calls(
        (
            call(
                Path("fake_path/index.html"),
                "http://example.com/index.html",
                False,
                ANY,
            ),
            call(
                Path("fake_path/fake_page_1.html"),
                "http://example.com/fake_page_1.html",
                False,
                ANY,
            ),
            call(
                Path("fake_path/fake_page_2.html"),
                "http://example.com/fake_page_2.html",
                False,
                ANY,
            ),
        )
    )


def test___crawl_download_errors(mocker: MockerFixture) -> None:
    """Test the Crawler.__crawl() method when downloading fails."""

    # This time we'll have everything but the first page fail to download.
    mock_download = mocker.patch.object(Crawler, "download_resource")
    mock_download.side_effect = [
        (Path("fake_path/index.html"), "http://example.com/index.html"),
        IOError("OMG we couldn't download anything!"),
    ]
    mock_get_urls = mocker.patch.object(Crawler, "get_new_domain_urls")
    mock_get_urls.side_effect = [
        ("http://example.com/fake_page_1.html",),
    ]

    crawler = Crawler()
    output_dir = Path("fake_path/")
    # pylint: disable-next=protected-access
    errors = crawler._Crawler__crawl("http://example.com/", output_dir)  # type: ignore

    assert errors == 1

    # We don't care how the code does it, just that tries to download each file once.
    assert mock_download.call_count == 2
    mock_download.assert_has_calls(
        (
            call("http://example.com/", output_dir),
            call("http://example.com/fake_page_1.html", output_dir),
        ),
        any_order=True,
    )

    # Simiarly, we don't care how the code does it, just that it tries to scan each file once.
    assert mock_get_urls.call_count == 1
    mock_get_urls.assert_has_calls(
        (
            call(
                Path("fake_path/index.html"),
                "http://example.com/index.html",
                False,
                ANY,
            ),
        )
    )


def test___crawl_scan_errors(mocker: MockerFixture) -> None:
    """Test the Crawler.__crawl() method when URL scanning fails."""

    # This time we'll have everything but the first page fail to scan.
    mock_download = mocker.patch.object(Crawler, "download_resource")
    mock_download.side_effect = [
        (Path("fake_path/index.html"), "http://example.com/index.html"),
    ]
    mock_get_urls = mocker.patch.object(Crawler, "get_new_domain_urls")
    mock_get_urls.side_effect = [IOError("OMG we couldn't scan anything!")]

    crawler = Crawler()
    output_dir = Path("fake_path/")
    # pylint: disable-next=protected-access
    errors = crawler._Crawler__crawl("http://example.com/", output_dir)  # type: ignore

    assert errors == 1

    # We don't care how the code does it, just that tries to download each file once.
    assert mock_download.call_count == 1
    mock_download.assert_has_calls(
        (call("http://example.com/", output_dir),),
        any_order=True,
    )

    # Simiarly, we don't care how the code does it, just that it tries to scan each file once.
    assert mock_get_urls.call_count == 1
    mock_get_urls.assert_has_calls(
        (
            call(
                Path("fake_path/index.html"),
                "http://example.com/index.html",
                False,
                ANY,
            ),
        )
    )


def test___crawl_multiple_errors(mocker: MockerFixture) -> None:
    """Test the Crawler.__crawl() method when things go well."""

    # In this case we have the index link to both page 1 and page 2, but page 1 fails to download
    # and page 2 downloads, but fails to scan.
    mock_download = mocker.patch.object(Crawler, "download_resource")
    mock_download.side_effect = [
        (Path("fake_path/index.html"), "http://example.com/index.html"),
        IOError("OMG we couldn't download page 1!"),
        (Path("fake_path/fake_page_2.html"), "http://example.com/fake_page_2.html"),
    ]
    mock_get_urls = mocker.patch.object(Crawler, "get_new_domain_urls")
    mock_get_urls.side_effect = [
        (
            "http://example.com/fake_page_1.html",
            "http://example.com/fake_page_2.html",
        ),
        IOError("OMG we couldn't scan page 2!"),
    ]

    crawler = Crawler()
    output_dir = Path("fake_path/")
    # pylint: disable-next=protected-access
    errors = crawler._Crawler__crawl("http://example.com/", output_dir)  # type: ignore

    assert errors == 2

    # We don't care how the code does it, just that tries to download each file once.
    assert mock_download.call_count == 3
    mock_download.assert_has_calls(
        (
            call("http://example.com/", output_dir),
            call("http://example.com/fake_page_1.html", output_dir),
            call("http://example.com/fake_page_2.html", output_dir),
        ),
        any_order=True,
    )

    # Simiarly, we don't care how the code does it, just that it tries to scan each file once.
    assert mock_get_urls.call_count == 2
    mock_get_urls.assert_has_calls(
        (
            call(
                Path("fake_path/index.html"),
                "http://example.com/index.html",
                False,
                ANY,
            ),
            call(
                Path("fake_path/fake_page_2.html"),
                "http://example.com/fake_page_2.html",
                False,
                ANY,
            ),
        )
    )
