"""test_crawler_crawl_method.py - Test the babyspider.crawler.Crawler.crawl() method."""

import logging
from pathlib import Path
from unittest.mock import MagicMock
from urllib.robotparser import RobotFileParser

import pytest
from pytest_mock import MockerFixture
from requests import RequestException

from babyspider.crawler import Crawler


def setup_mocks(mocker: MockerFixture) -> tuple[MagicMock, MagicMock]:
    """Setup the mocks to for testing Crawler.crawl()."""

    # Setup the Crawler method mocks.
    mock_dl_sitemaps = mocker.patch.object(Crawler, "download_sitemap_urls", return_value=0)
    mock___crawl = mocker.patch.object(Crawler, "_Crawler__crawl", return_value=0)

    # This lets us not bother with mocking more of a robots.txt file.
    mocker.patch.object(Crawler, "_Crawler__get_minimum_crawl_delay", return_value=3)
    # This has its own test and saves us from mocking HTTP calls.
    mocker.patch.object(
        Crawler,
        "_get_starting_and_robots_urls",
        return_value=("https://example.com/", "https://example.com/robots.txt"),
    )
    # This stops us from making directories on disk and helps us check when we add timestamps.
    mocker.patch.object(
        Crawler,
        "_make_output_dir",
        lambda self, output_dir_name, add_timestamp: Path(f"{output_dir_name}_{add_timestamp}"),
    )
    # This has its own test and saves us from mocking HTTP calls.
    mocker.patch.object(Crawler, "get_canonical_url", return_value="https://example.com/")

    # Setup the RobotFileParser method mocks.
    # We'll simulate a robots.txt file so we only test the crawl method.
    mocker.patch.object(RobotFileParser, "can_fetch", return_value=True)
    mocker.patch.object(RobotFileParser, "read")
    mocker.patch.object(
        RobotFileParser, "site_maps", return_value=["https://example.com/sitemap.xml"]
    )

    return mock_dl_sitemaps, mock___crawl


def test_crawl_happy_path(mocker: MockerFixture, caplog: pytest.LogCaptureFixture) -> None:
    """Test the Crawler.crawl() method with when things go as expected."""
    # This should create a directory name based on the starting URL domain with a timestamp.
    # mock_start_urls, mock_canon_url, mock_rp_read, mock_rp_can_fetch = setup_mocks(mocker)
    mock_dl_sitemaps, mock___crawl = setup_mocks(mocker)

    # The setup arguments are lists of tuples of this format:
    #
    #   (Crawler_kwargs, [output_dir_name, images, add_timestamp])
    #
    # The type hint for this is rather hideous and more confusing that it's worth. So, instead
    # we'll just disable typing here, and then check the types of each item in the loop.
    setup_args = (  # type: ignore
        # Test the default arguments.
        ({}, [None, None, None]),
        # Test the without a timestamp.
        ({}, [None, None, False]),
        # Test specifying and output directory.
        ({}, ["fake_dir", None, None]),
        # Test crawling images.
        ({}, [None, True, None]),
        # Test specifying everything.
        ({}, ["fake_dir", True, True]),
        # Test with no sitemaps and only crawling.
        ({"use_sitemaps": "never"}, [None, None, None]),
        # Test with only sitemaps and no crawling.
        ({"use_sitemaps": "only"}, [None, None, None]),
    )

    for crawler_kwargs, crawl_args in setup_args:
        # Check our fixture data since using type annotations it too painful.
        assert isinstance(
            crawler_kwargs, dict
        ), f"Bad crawler kwargs fixture data: {crawler_kwargs!r}"

        crawler = Crawler(**crawler_kwargs)
        output_dir_name, images, add_timestamp = crawl_args

        # Check our fixture data since using type annotations it too painful.
        assert output_dir_name is None or isinstance(
            output_dir_name, str
        ), f"Bad output directory name fixture data: {output_dir_name!r}"
        assert images is None or isinstance(images, bool), f"Bad images fixture data: {images!r}"
        assert add_timestamp is None or isinstance(
            add_timestamp, bool
        ), f"Bad add timestamp fixture data: {add_timestamp!r}"

        # The lambda function we use to mock _make_output_dir returns a final directory Path of the
        # output_dir_name and add_timestamp passed in.
        # The business logic says we should use the domain name if no directory name is given and
        # if we do that, also add a timestamp by default.
        if not output_dir_name:
            if add_timestamp is None:
                expected_output_path = Path("example.com_True")
            else:
                expected_output_path = Path(f"example.com_{add_timestamp}")
        else:
            expected_output_path = Path(f"{output_dir_name}_{add_timestamp}")

        output_path, starting_url, errors = crawler.crawl(
            "example.com",
            output_dir_name=output_dir_name,
            # Having images be None here is part of testing.
            images=images,  # type: ignore
            add_timestamp=add_timestamp,
        )

        if crawler_kwargs.get("use_sitemaps") == "never":
            mock_dl_sitemaps.assert_not_called()
        else:
            mock_dl_sitemaps.assert_called_once_with(
                "https://example.com/sitemap.xml", expected_output_path
            )

        if crawler_kwargs.get("use_sitemaps") == "only":
            mock___crawl.assert_not_called()
        else:
            mock___crawl.assert_called_once_with(
                "https://example.com/", expected_output_path, images, 0
            )
        assert output_path == str(expected_output_path)
        assert starting_url == "https://example.com/"
        assert errors == 0

        # Make sure we get warned about mixing image crawling with only downloading sitemaps when
        # and only when needed.
        if images and crawler_kwargs.get("use_sitemaps") == "only":
            assert (
                "Crawling for images is set, but we're not crawling for URLs at all, only "
                "downloading from sitemaps!"
                in [record.msg for record in caplog.records if record.levelno == logging.WARNING]
            ), (
                "A warning should be logged when image crawling is set but we are only "
                "downloading from sitemaps."
            )
        else:
            assert (
                "Crawling for images is set, but we're not crawling for URLs at all, only "
                "downloading from sitemaps!" not in caplog.messages
            ), (
                "A log message about image crawling when we are only downloading from "
                "sitemaps was shown when image crawling was not set."
            )

        caplog.clear()
        mock_dl_sitemaps.reset_mock()
        mock___crawl.reset_mock()


def test_crawl_counting_errors(mocker: MockerFixture) -> None:
    """Test the Crawler.crawl() method returns total errors."""
    setup_mocks(mocker)
    mock_dl_sitemaps = mocker.patch.object(Crawler, "download_sitemap_urls", return_value=2)
    mock___crawl = mocker.patch.object(Crawler, "_Crawler__crawl", return_value=3)

    crawler = Crawler()
    output_path, starting_url, errors = crawler.crawl("example.com")

    mock_dl_sitemaps.assert_called_once_with(
        "https://example.com/sitemap.xml", Path("example.com_True")
    )
    mock___crawl.assert_called_once_with("https://example.com/", Path("example.com_True"), False, 2)
    assert output_path == "example.com_True"
    assert starting_url == "https://example.com/"
    assert errors == 5


def test_crawl_no_sitemaps(mocker: MockerFixture) -> None:
    """Test the Crawler.crawl() method handles when there are no sitemaps."""
    mock_dl_sitemaps, mock___crawl = setup_mocks(mocker)
    mocker.patch.object(RobotFileParser, "site_maps", return_value=None)

    crawler = Crawler()
    output_path, starting_url, errors = crawler.crawl("example.com")

    mock_dl_sitemaps.assert_not_called()
    mock___crawl.assert_called_once_with("https://example.com/", Path("example.com_True"), False, 0)
    assert output_path == "example.com_True"
    assert starting_url == "https://example.com/"
    assert errors == 0


def test_crawl_no_sitemaps_but_sitemaps_only(mocker: MockerFixture) -> None:
    """Test the Crawler.crawl() handles getting no sitemaps when only downloading from sitemaps."""
    mock_dl_sitemaps, mock___crawl = setup_mocks(mocker)
    mocker.patch.object(RobotFileParser, "site_maps", return_value=None)

    crawler = Crawler(use_sitemaps="only")
    with pytest.raises(ValueError):
        crawler.crawl("example.com")

    mock_dl_sitemaps.assert_not_called()
    mock___crawl.assert_not_called()


def test_crawl_skip_forbidden_url(mocker: MockerFixture) -> None:
    """Test the Crawler.crawl() won't read a starting URL forbidden by robots.txt."""

    mock_dl_sitemaps, mock___crawl = setup_mocks(mocker)
    mocker.patch.object(RobotFileParser, "can_fetch", return_value=False)

    crawler = Crawler()
    with pytest.raises(ValueError):
        crawler.crawl("example.com")

    mock_dl_sitemaps.assert_not_called()
    mock___crawl.assert_not_called()


def test_crawl_bad_canonincal_url(mocker: MockerFixture) -> None:
    """Test the Crawler.crawl() handles errors getting the canonical version of the starting URL."""

    mock_dl_sitemaps, mock___crawl = setup_mocks(mocker)
    mocker.patch.object(
        Crawler,
        "get_canonical_url",
        side_effect=RequestException("Fake too many redirects error!"),
    )

    crawler = Crawler()
    with pytest.raises(ValueError):
        crawler.crawl("example.com")

    mock_dl_sitemaps.assert_not_called()
    mock___crawl.assert_not_called()
