"""test_crawler_nonpublic_methods.py - Test babyspider.crawler.Crawler proteced and private methods,
except __crawl().

See test_crawler___crawl_method.py for tests for babyspider.crawler.Crawler.__crawl().
"""

from datetime import datetime
from pathlib import Path
from urllib.robotparser import RobotFileParser

import pytest
from pytest_mock import MockerFixture

from babyspider.crawler import Crawler


def test__get_starting_and_robots_urls() -> None:
    """Test the Crawler._get_starting_and_robots_urls() method."""

    given_and_expected = (
        ("example.com", ("https://example.com", "https://example.com/robots.txt")),
        (
            "https://example.com",
            ("https://example.com", "https://example.com/robots.txt"),
        ),
        (
            "https://example.com/;a=1&b=2?c=3&d=4#five",
            ("https://example.com/", "https://example.com/robots.txt"),
        ),
        (
            "example.com/;a=1&b=2?c=3&d=4#five",
            ("https://example.com/", "https://example.com/robots.txt"),
        ),
        ("http://example.com", ("http://example.com", "http://example.com/robots.txt")),
        (123, ValueError),
        (123.4, ValueError),
        ("file:///pickles.txt", ValueError),
    )

    crawler = Crawler()

    for given, expected in given_and_expected:
        if isinstance(expected, type) and issubclass(expected, BaseException):
            try:
                with pytest.raises(expected):
                    # We ignore type here because the whole point is to test handling wrong types.
                    # pylint: disable-next=protected-access
                    results = crawler._get_starting_and_robots_urls(given)  # type: ignore
            except AssertionError:
                pytest.fail(
                    f"Crawler._get_starting_and_robots_urls({given!r}) should raise {expected!r} "
                    f"but insteade returned {results!r}!"
                )
        else:
            results = crawler._get_starting_and_robots_urls(  # pylint: disable=protected-access
                # We ignore type here because the whole point is to test handling wrong types.
                given  # type: ignore
            )
            assert results == expected, (
                f"Crawler._get_starting_and_robots_urls({given!r}) should return {expected!r} but "
                f"not {results!r}!"
            )


def test__make_output_dir(mocker: MockerFixture) -> None:
    """Test the Crawler._make_output_dir() method."""
    mocker.patch("babyspider.crawler.datetime").now.return_value = datetime(
        2025, 12, 25, 12, 34, 56
    )
    mock_exists = mocker.patch.object(Path, "exists")
    mock_is_dir = mocker.patch.object(Path, "is_dir")
    mock_mkdir = mocker.patch.object(Path, "mkdir")

    crawler = Crawler("fake_base_dir")

    mock_exists.return_value = False
    # Test creating the directory without adding timestamp with a string path.
    result_path = crawler._make_output_dir("subdir", False)  # pylint: disable=protected-access
    assert result_path == Path("fake_base_dir/subdir")
    mock_mkdir.assert_called_once_with(parents=True)
    mock_mkdir.reset_mock()

    # Test creating the directory with adding a timestamp with a Path object.
    result_path = crawler._make_output_dir(Path("subdir"), True)  # pylint: disable=protected-access
    assert result_path == Path("fake_base_dir/subdir_20251225_1234")
    mock_mkdir.assert_called_once_with(parents=True)
    mock_mkdir.reset_mock()

    # Test failing to create the directory path due to type issues
    with pytest.raises(TypeError):
        # We ignore type here because using a bad type is the point of the test.
        # pylint: disable-next=protected-access
        crawler._make_output_dir(None, True)  # type: ignore
    mock_mkdir.assert_not_called()

    mock_exists.return_value = True
    mock_is_dir.return_value = True
    # Test the target directory already existing.
    result_path = crawler._make_output_dir(  # pylint: disable=protected-access
        Path("subdir"), False
    )
    assert result_path == Path("fake_base_dir/subdir")
    mock_mkdir.assert_not_called()

    mock_is_dir.return_value = False
    # Test the target directory already existing but its not a directory.
    with pytest.raises(ValueError):
        crawler._make_output_dir("subdir", True)  # pylint: disable=protected-access
    mock_mkdir.assert_not_called()


def test_crawler___init__() -> None:
    """Test the Crawler.__init__ method."""
    crawler = Crawler("my_dir", "html.parser", 5, "never")
    assert (
        crawler.base_output_dir == Path("my_dir")
        and crawler.parser == "html.parser"
        and crawler.crawl_delay == 5
        and crawler.use_sitemaps == "never"
    ), "Crawler('my_dir', 'html.parser', 5, 'never') did not set the correct public atributes."
    # pylint: disable=protected-access
    assert (
        # Sadly, mypy doesn't understand private attribute name mangling.
        crawler._Crawler__known_urls == set()  # type: ignore
        and isinstance(crawler._Crawler__robot_parser, RobotFileParser)  # type: ignore
        and crawler._Crawler__current_crawl_delay == 5  # type: ignore
    ), "Crawler('my_dir', 'html.parser', 5, 'never') did not set the correct private atributes."
    # pylint: enable=protected-access
    crawler = Crawler(use_sitemaps="invalid")
    assert crawler.use_sitemaps == "first", (
        "Callling Crawler() with use_sitemaps other than 'first', 'never', or 'only' should set "
        "it to 'first'"
    )
    crawler = Crawler(use_sitemaps=123)  # type: ignore
    assert crawler.use_sitemaps == "first", (
        "Callling Crawler() with use_sitemaps other than 'first', 'never', or 'only' should set "
        "it to 'first'"
    )
    crawler = Crawler(crawl_delay=-5)
    assert crawler.crawl_delay == 0, (
        "Crawler(crawl_delay=-5) or other negative numbers should prevent attempted time travel "
        "and set 'crawl_delay' to 0."
    )

    with pytest.raises(ValueError):
        Crawler(parser="fake_parser")


def test_crawler___str__() -> None:
    """Test the Crawler string method."""
    crawler_with_dir = str(Crawler("my_favorite_directory"))
    assert (
        "Crawler" in crawler_with_dir
    ), "String versions of Crawler() should have the class name in it."
    assert (
        "my_favorite_directory" in crawler_with_dir
    ), "String versions of Crawler() should show the base output directory."


def test_crawler___repr__() -> None:
    """Test the Crawler representational string method."""
    assert repr(Crawler("my_dir", "html.parser", 5, "never")) == (
        f"<Crawler(base_output_dir={Path('my_dir')!r}, parser='html.parser', "
        "crawl_delay=5, use_sitemaps='never')>"
    ), (
        "The represenational string version of Crawler should be a Python literal of the call to "
        "create the object enclosed in angle brackets."
    )
