"""test_crawler_public_methods.py - Test the babyspider.crawler.Crawler public instance methods,
except crawl().

See test_crawler_static_methods.py for static instance methods.
See test_crawler_crawl_method.py for babyspider.crawler.Crawler.crawl().
"""

import tempfile
from io import BytesIO
from pathlib import Path
from urllib.robotparser import RobotFileParser

import bs4
import pytest
from pytest_mock import MockerFixture
from requests import RequestException, Response

from babyspider.crawler import Crawler

FIXTURE_PATH = Path(__file__).parent / "fixtures"


def test_download_resource(mocker: MockerFixture) -> None:
    """Test the Crawler.download_resource() method."""

    mock_get_url = mocker.patch.object(Crawler, "get_canonical_url")
    mock_req_get = mocker.patch("requests.get")
    mock_sleep = mocker.patch("time.sleep")

    # In case something goes wrong, we'll run everything with crawler that uses a temporary
    # directory for file access.
    with tempfile.TemporaryDirectory(prefix="babyspider_test_") as tempdir:
        crawler = Crawler(tempdir)

        # Check handling of a failed attempt to get the canonical URL.
        mock_get_url.side_effect = RequestException("Fake forbidden URL error!")
        with pytest.raises(IOError):
            crawler.download_resource("http://example.com/file.html", tempdir)

        # Check handing being redirected outside the domain.
        mock_get_url.side_effect = lambda url: "http://somewhereelse.com/file.html"
        with pytest.raises(ValueError):
            crawler.download_resource("http://example.com/file.html", tempdir)

        # Check file is downloading, with any needed subdirectories made beforehand,  and that we
        # wait the configured crawl delay.
        mock_get_url.side_effect = lambda url: "http://example.com/subdir/final_file.html"

        # Setup our failed download response.
        response = Response()
        response.url = "http://example.com/final_file.html"
        response.status_code = 404
        response._content = b"404 Not Found"  # pylint: disable=protected-access
        response.raw = BytesIO(response._content)  # pylint: disable=protected-access
        response.reason = "Not Found"
        response.encoding = "utf-8"
        response.headers["Content-Type"] = "text/html; charset=utf-8"
        # We're ignoring type here because while this sort of monkeypatching is usually bad idea,
        # it's the best of the bad options in mocking things for tests.
        response.__enter__ = lambda self=response: self  # type: ignore
        response.__exit__ = lambda self, exc_type, exc_val, exc_tb: False  # type: ignore
        mock_req_get.return_value = response

        # Check handling when the download fails.
        with pytest.raises(RequestException):
            crawler.download_resource("http://example.com/bad_file.html", tempdir)

        # Setup our successful download response.
        with open(str(FIXTURE_PATH / "fixture_file.html"), "rt", encoding="utf-8") as handle:
            source_file_content = handle.read()
        response._content = source_file_content.encode("utf-8")  # pylint: disable=protected-access
        response.raw = BytesIO(response._content)  # pylint: disable=protected-access
        response.status_code = 200
        response.url = "http://example.com/final_file.html"
        response.reason = "OK"

        # Check successfully downloading the file.
        local_path, final_url = crawler.download_resource("http://example.com/file.html", tempdir)
        assert local_path == Path(f"{tempdir}/subdir/final_file.html")
        assert final_url == "http://example.com/subdir/final_file.html"
        mock_sleep.assert_called_with(crawler.crawl_delay)
        assert local_path.exists()
        with open(str(local_path), "rt", encoding="utf-8") as handle:
            downloaded_file_content = handle.read()
        assert (
            downloaded_file_content == source_file_content
        ), "Calling Crawler.download_resource() mangled the file contents it downloaded!"

        # Check skipping download when the file is already present.
        mock_req_get.reset_mock()
        local_path, final_url = crawler.download_resource("http://example.com/file.html", tempdir)
        mock_req_get.assert_not_called()

        # Check raising an exception if the target download file already exists as a directory.
        local_path.unlink()
        local_path.mkdir()
        with pytest.raises(ValueError):
            local_path, final_url = crawler.download_resource(
                "http://example.com/file.html", tempdir
            )


def test_download_sitemap_urls(mocker: MockerFixture) -> None:
    """Test the Crawler.download_sitemap_urls() method."""

    # Setup all our mocks before we create the crawler.
    mock_get = mocker.patch(
        "requests.get",
    )
    mock_get_urls = mocker.patch.object(Crawler, "get_urls_from_sitemap")
    mock_dl_resource = mocker.patch.object(
        Crawler, "download_resource", side_effect=lambda url, output_dir: (None, url)
    )

    crawler = Crawler()

    # Check handling problems downloading the sitemap itself.
    # This should return 1 as a single error.
    mock_get.side_effect = RequestException("Fake sitemap download error!")
    assert crawler.download_sitemap_urls("http://example.com/sitemap.xml", "fake_path") == 1
    mock_get.side_effect = None

    # Check downloading working URLs.
    # pylint: disable-next=protected-access
    crawler._Crawler__known_urls = set()  # type: ignore
    fake_found_urls = {
        "http://example.com/index.html",
        "http://example.com/about_bad.html",
        "http://example.com/contact_bad.html",
        "http://example.com/shop.html",
    }
    mock_get_urls.return_value = fake_found_urls
    errors = crawler.download_sitemap_urls("http://example.com/sitemap.xml", "fake_path")
    assert errors == 0, (
        "Calling crawler.download_sitemap_urls() with no bad files should have returned 0 errors, "
        f"not {errors}!"
    )
    assert (
        # pylint: disable-next=protected-access
        crawler._Crawler__known_urls == fake_found_urls  # type: ignore
    ), "Calling crawler.download_sitemap_urls() should capture all URLs to the known set!"
    assert mock_dl_resource.call_count == len(
        fake_found_urls
    ), "Calling crawler.download_sitemap_urls() should attempt to download all found URLs!"

    # Check counting errors when downloading URLs.
    mock_dl_resource.reset_mock()
    # pylint: disable-next=protected-access
    crawler._Crawler__known_urls = set()  # type: ignore
    fake_found_urls = {
        "http://example.com/index.html",
        "http://example.com/about_bad.html",
        "http://example.com/contact_bad.html",
        "http://example.com/shop.html",
    }
    mock_get_urls.return_value = fake_found_urls

    def fake_dl_resource(url: str, output_dir: str | Path) -> tuple[Path, str]:
        """Fake varying behavior in Crawler.download_resource() calls."""
        if "bad" in url:
            raise IOError("Can't download URLs with the 'bad' in the name!")
        return Path(output_dir) / url.split("/")[-1], url

    mock_dl_resource.side_effect = fake_dl_resource
    errors = crawler.download_sitemap_urls("http://example.com/sitemap.xml", "fake_path")
    assert errors == 2, (
        "Calling crawler.download_sitemap_urls() with 2 bad files should have returned 2 errors, "
        f"not {errors}!"
    )
    assert (
        # pylint: disable-next=protected-access
        crawler._Crawler__known_urls == fake_found_urls  # type: ignore
    ), "Calling crawler.download_sitemap_urls() should capture all URLs to the known set!"
    assert mock_dl_resource.call_count == len(
        fake_found_urls
    ), "Calling crawler.download_sitemap_urls() should attempt to download all found URLs!"


def test_get_canonical_url(mocker: MockerFixture) -> None:
    """Test the Crawler.get_canonical_url() method."""

    # Setup all the mocks before creating our crawler.
    mock_can_fetch = mocker.patch.object(RobotFileParser, "can_fetch", return_value=True)
    mocker.patch.object(RobotFileParser, "read")
    mocker.patch.object(RobotFileParser, "site_maps", return_value=[])
    mock_head = mocker.patch("requests.head")

    crawler = Crawler()

    # Test that we don't pull URLs forbidden by robots.txt.
    mock_can_fetch.return_value = False
    with pytest.raises(RequestException):
        crawler.get_canonical_url("http://example.com/forbidden.html")
    mock_can_fetch.return_value = True

    # Test that we give up after too many redirects.
    mock_head.reset_mock()
    fake_redirect = Response()
    fake_redirect.status_code = 302
    fake_redirect.url = "http://example.com/loop"
    fake_redirect.headers["Location"] = "http://example.com/loop"
    fake_redirect.reason = "Found"
    mock_head.return_value = fake_redirect
    with pytest.raises(RequestException):
        crawler.get_canonical_url("http://example.com/loop")
    # There will be one more call to requests.head() for the initial check before th redirects
    assert mock_head.call_count == crawler.max_redirects + 1

    # Test a reasonable redirect with a relative URL, and make sure we don't drop any user,
    # password, port number, params, query, or frament components.
    mock_head.reset_mock()

    # Use a relative path for the redirect this time.
    fake_redirect.url = "/old_version"
    fake_redirect.headers["Location"] = "/old_version"
    fake_result = Response()
    fake_result.status_code = 200
    fake_result.url = "/new_version;a=1&b=2?c=3&d=4#fragment"
    fake_result.headers["Location"] = "/new_version;a=1&b=2?c=3&d=4#fragment"
    fake_result.reason = "Found"
    mock_head.side_effect = [fake_redirect, fake_result]
    result_url = crawler.get_canonical_url("http://user:pass@example.com:80/old_version")
    assert result_url == "http://user:pass@example.com:80/new_version;a=1&b=2?c=3&d=4#fragment"


def test_get_local_path() -> None:
    """Test the Crawler.get_local_path() method."""
    crawler = Crawler()

    # Note that method returns Path objects, but for the fixture data here we just use strings and
    # turn them into Path objects in the loop below to keep this simpler.
    given_and_expected = (
        ("https://example.com/file.html", "file.html"),
        ("https://example.com/dir/file.html", "dir/file.html"),
        ("https://example.com/dir/subdir/file.html", "dir/subdir/file.html"),
        ("https://example.com/", crawler.index_filename),
        ("https://example.com/dir/", f"dir/{crawler.index_filename}"),
        ("https://example.com/dir/subdir/", f"dir/subdir/{crawler.index_filename}"),
        ("https://example.com/file.html;a=1&b=2", "file.html"),
        ("https://example.com/file.html?c=3&d=4", "file.html"),
        ("https://example.com/file.html#frangment", "file.html"),
        ("https://example.com/file.html;a=1&b=2?c=3&d=4#frangment", "file.html"),
        ("https://example.com/;a=1&b=2", crawler.index_filename),
        ("https://example.com/?c=3&d=4", crawler.index_filename),
        ("https://example.com/#frangment", crawler.index_filename),
        ("https://example.com/;a=1&b=2?c=3&d=4#frangment", crawler.index_filename),
    )
    for given, expected_str in given_and_expected:
        expected = Path(expected_str)
        result = crawler.get_local_path(given)
        assert (
            result == expected
        ), f"Calling Crawler.get_local_path({given!r}) should return {expected!r} not {result!r}!"

    # Check handling of unparsable URLs.
    with pytest.raises(ValueError):
        crawler.get_local_path(123)  # type: ignore


def test_get_new_domain_urls(mocker: MockerFixture) -> None:
    """Test the Crawler.get_new_domain_urls() method."""
    mock_soup = mocker.patch("bs4.BeautifulSoup").return_value
    crawler = Crawler()

    # Check handling general issues reading files from disk.
    mock_soup.side_effect = IOError("Fake 'you can't read this file' error!")
    try:
        with pytest.raises(IOError):
            crawler.get_new_domain_urls(Path("fake_path/bad_file"), "http://example.com/bad_file")
    except AssertionError:
        pytest.fail("Calling get_new_domain_urls() ignored IOErrors when reading a bad file.")

    # Check handling non-HTML files (or anything BeautifulSoup can't parse).
    mock_soup.side_effect = AttributeError
    try:
        with pytest.raises(IOError):
            crawler.get_new_domain_urls(Path("fake_path/file.png"), "http://example.com/file.png")
    except AssertionError:
        pytest.fail(
            "Calling get_new_domain_urls() on ignored parsing errors when reading an image file."
        )

    # Check general URL extraction.

    # The format of items in the given, fixture, and expected tuple is:
    #
    #   (get_new_domain_urls_args, fixture_file, expected_urls_set)
    #
    given_fixture_and_expected = (
        (
            (
                "example.com/fake_file.html",
                "https://example.com/fake_file.html",
                False,
                None,
            ),
            "fixture_file.html",
            set(["https://example.com/pickles.html", "http://example.com/tina.html"]),
        ),
        (
            (
                "example.com/fake_file.html",
                "https://example.com/fake_file.html",
                True,
                None,
            ),
            "fixture_file.html",
            set(
                [
                    "https://example.com/pickles.html",
                    "http://example.com/tina.html",
                    "https://example.com/image1.png",
                    "https://example.com/image2.png",
                ]
            ),
        ),
        (
            (
                "example.com/fake_file.html",
                "https://example.com/fake_file.html",
                True,
                ["http://example.com/tina.html", "https://example.com/image2.png"],
            ),
            "fixture_file.html",
            set(["https://example.com/pickles.html", "https://example.com/image1.png"]),
        ),
    )
    for given, fixture, expected in given_fixture_and_expected:
        fixture_path = str(FIXTURE_PATH / fixture)
        with open(fixture_path, "rt", encoding="utf-8") as handle:
            mocker.patch("builtins.open", mocker.mock_open(read_data=handle.read()))
        results = crawler.get_new_domain_urls(*given)
        assert results == expected, (
            f"Calling get_new_domain_urls({given!r}) backed by fixture file "
            f"{fixture_path!r} should have returned {expected!r} not {results!r}!"
        )


def test_get_urls_from_sitemap(mocker: MockerFixture) -> None:
    """Test the Crawler.get_urls_from_sitemap() method."""
    crawler = Crawler()

    # Check handling a non-parsable sitemap.
    mock_soup = mocker.patch("bs4.BeautifulSoup").return_value
    mock_soup.side_effect = UnicodeDecodeError
    assert crawler.get_urls_from_sitemap(b"Fake unparsable content!") == set()
    mock_soup.stop()

    # Check falling back to an HTML parser when an XML one is not available.
    def fake_bs_init(content: bytes, parser: str) -> None:
        """A fake BeautifulSoup.__init__() that raises FeatureNotFound if the parser is 'xml' and
        raises a ValueError for everything else."""
        if parser == "xml":
            raise bs4.FeatureNotFound(f"Fake no XML parser error for {content!r}!")
        raise ValueError(f"Fake non-XML parser error for {content!r}!")

    mock_soup = mocker.patch("bs4.BeautifulSoup", side_effect=fake_bs_init)
    crawler.parser = "xml"
    with pytest.raises(ValueError):
        crawler.get_urls_from_sitemap(b"fake sitemap content")
    crawler.parser = "html.parser"
    mock_soup.stop()

    # Othe than the parser fallback logic, this is just a simple filter on a BeautifulSoup tag list.
    # Given that, we only need one small test fixture for the parsing logic.
    example_sitemap = b"""
    <?xml version="1.0" encoding="UTF-8"?>
    <urlset
        xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">

      <url>
        <loc>https://www.example.com/</loc>
        <lastmod>2025-05-17</lastmod>
        <changefreq>weekly</changefreq>
        <priority>1.0</priority>
      </url>

      <url>
        <loc>https://www.example.com/about</loc>
        <lastmod>2025-05-10</lastmod>
        <changefreq>monthly</changefreq>
        <priority>0.8</priority>
      </url>

      <url>
        <loc>https://www.example.com/contact</loc>
        <lastmod>2025-04-28</lastmod>
        <changefreq>yearly</changefreq>
        <priority>0.5</priority>
      </url>

    </urlset>
    """
    assert crawler.get_urls_from_sitemap(example_sitemap) == {
        "https://www.example.com/",
        "https://www.example.com/about",
        "https://www.example.com/contact",
    }


def test_read_robots_txt(mocker: MockerFixture) -> None:
    """Test the Crawler.read_robots_txt() method."""

    # Test that we setup a new robot.txt parser and call read() on it.
    class FakeRobotFileParser(RobotFileParser):
        """A fake RobotFileParser for testing."""

        def read(self) -> None:
            """Don't actually try to get the robots.txt file."""

    fake_rp = FakeRobotFileParser()
    mock_rp = mocker.patch("babyspider.crawler.RobotFileParser", return_value=fake_rp)
    crawler = Crawler()
    mock_rp.reset_mock()
    crawler.read_robots_txt("https://example.com/robots.txt")
    # pylint: disable=protected-access
    assert crawler._Crawler__robot_parser == fake_rp, (  # type: ignore
        "Calling Crawler.read_robots_txt() did not set self.__robot_file_parser to a a new "
        "urllib.robotparser.RobotFileParser, but instead: "
        f"{crawler._Crawler__robot_parser!r}"  # type: ignore
    )
    # pylint: enable=protected-access
    mock_rp.assert_called_once()

    # Test that we handle errors reading the robots.txt file.
    mock_rp.side_effect = ValueError
    with pytest.raises(ValueError):
        crawler.read_robots_txt("https://example.com/robots.txt")
